/************************************************************************************************************************/
/*    

Programmer:  Klaus Schneider, Vision Components (www.vision-components.de)
Date:        29.11.2005


Demo program for register mark inspection:
------------------------------------------

Download the demo program to any TI based Vision Components smart camera like VC20xx, VC4xxx, and so on.
Then start the program. 

After starting the program, you can press the key "F1" in order to display the camera menu. 
Here you can change all setting parameters like threshold, search area, shutter time and so on.

For example - Menu Options:

--------------------------------------------------------------------------------
l:   live-picture
s:   change shutter time
t:   change threshold
a:   change search area
x:   display results

q:   quit Menu
--------------------------------------------------------------------------------

Before you quit the menu, the program will ask you, if you want to store the new values (like search area)
on the Flash so that the program starts with the new values at the next time. If you select 'n' for no, you 
can use the new values only temporally. Next time, when you start the program again, you use the old values.

For example:

Save New Parameters (y/n)?



If you like to start the image calculation, you have three options to trigger the program:

1. with an external trigger, connected to the high speed camera trigger input
2. by pressing the Key "ENTER" at the VC Keypad
3. sending "ENTER" over a terminal program 

For example: Start your terminal program and connect it to the camera. Then press "ENTER" (option 3).


The camera will do a single calculation of the last image and displays the results on the video output.


The image calculation includes the following steps:

- the program use the function maskx() in order to binaries the search area and to be independent of any back light conditions.
- then it detects all blobs in the search area. 
- it eliminates all blobs, which are touching the border of the search area
- it sorts the remaining blobs by size
- then it takes the biggest blobs and sort them by x position
- The left blob (color black) is used as a reference.   


--------------------------------------------

Version 11:
-----------



*/
/************************************************************************************************************************/

// #define NEW_IMAGE_VAR

/************************************************************************************************************************/

#include <vcrt.h>
#include <vclib.h>
#include <macros.h>
#include <sysvar.h>
#include <flib.h>

#include <float.h>
#include <math.h>

/************************************************************************************************************************/

#define COLBLACK   8
#define COLWHITE  16
#define COLBLUE   24
#define COLRED    32
#define COLGREEN  40
#define COLGREY   48
#define COLCYAN   56
#define COLMAGE   64
#define COLDARK   72
#define COLYELLOW 80 

#define CompName "Vision Components"
#define PrgName  "Passermarken Version 10"

#define GOOD        0
#define BAD         1

#define MAX_MARKER  4

/***************************************************************************************************************************/

typedef struct
 {
  // define search area
  I32   SearchX;
  I32   SearchY;
  I32   SearchDx;
  I32   SearchDy;


  I32   Threshold;
  I32   ShowDetection;

  I32   StartShutter;

  I32   Error;

  I32   Ms;
  image Time;
 } 

Parameter;

/************************************************************************************************************************/

void IniCamera            (Parameter *Par);
void IniOverlayLUT        (void);
void CapturePosition      (int x, int y);
void IniParameter         (Parameter *Par);
void IniCameraBack        (Parameter *Par);

void DisplayMenu          (Parameter *Par);
void InvertValue          (I32 *ParValue, I8 *text);
void ChangeValue          (I32 *ParValue, I32 Min, I32 Max, I32 Step, I8 *text);

void DisplayResults       (Parameter *Par);
void SaveNewParameter     (Parameter *Par);
int  ReadParameter        (Parameter *Par);
int  StoreParameter       (Parameter *Par);
void LiveImage            (Parameter *Par);

void Low                  (image *area);
void High                 (image *area);

int  MeasureObject        (Parameter *Par);
I32  FindMarkers          (Parameter *Par, I32 threshold, ftr *f, I32 ShowDetection);
void SortStructure        (I32 *StructNr, I32 number, U32 *SortValue, I32 StructureSize);

void SelectArea           (int *px, int *py, int *pdx, int *pdy, int xmin, int xmax, int ymin, int ymax);

int  ReadFile             (char *name, char *data, int bytes);
int  StoreFile            (char *name, char *data, int bytes);
void ChangeColor          (image *area, int start, int end, int newcol);

/************************************************************************************************************************/

void main (void)

{
  int response=-1;
  Parameter  Par[1];

  IniCamera(Par);

  // first initialisation (try to read parameter from flash)
  IniParameter(Par);

  // select start shutter
  shutter(Par->StartShutter);

  // set trigger input for external triggered image acqusition (rising edge) 
  TRIGINP_POS();     

  // set trigger output to exposure controlled mode
  TRIGOUT_EXP();      
  TRIGOUT_POS();   
  
  print("Press ENTER to trigger program\n");                       
  print("Press 'F1'  for menu\n");                       
  print("Press ESC   to quit program\n");                       

  do
  {
START:

    // now start external triggered image acquisition (CCD sensor has to be in Still mode!)
    vmode(vmOvlStill);
    while(tenable() < 0);        
  
    // wait for trigger and image ready 
    while(!trdy())
    {
	  // does someone hit a key ?
      if ( kbhit() )
      {
	    // cancel external trigger image acquisition 
	    while(cancel_capture_rq());

        response = ReadKeyboard(0); 

        if (response == KEY_ENTER) {tpict(); goto MEASURE;}
        if (response == KEY_ESC  ) {goto QUIT;}
        if (response == KEY_F1   ) {DisplayMenu(Par); goto START;}
      }
    }

MEASURE:

    /********************************************************/
    /*  do calculation and measure time                     */
    /********************************************************/

    Par->Ms = getvar(MSEC);

    MeasureObject(Par);

    Par->Ms = getvar(MSEC)-Par->Ms;
    if(Par->Ms<0) Par->Ms+=1000;

    DisplayResults(Par);
  }
  while(1);

QUIT:

  IniCameraBack(Par);
}

/*************************************************************************************************************************/

void IniCamera  (Parameter *Par)

{
  I32   ScrX, ScrY, ScrDx, ScrDy;
  image Ovl;

  /************************/
  /* Define Overlay Color */
  /************************/

  IniOverlayLUT();

  /****************/
  /* Define Pages */
  /****************/

  ScrSetLogPage((I32)ScrGetDispPage);
  OvlSetLogPage((I32)OvlGetPhysPage);


  /* Define screen position in the center. X-position of sensor image must be in allignment of 16  */
  ScrDx = ScrGetColumns;
  ScrDy = ScrGetRows;
  ScrX  = ((DispGetColumns-ScrDx)/32)*16;
  ScrY  = ((DispGetRows   -ScrDy)/32)*16;

  /* Position of the image */
  CapturePosition(ScrX, ScrY);


  /*****************/
  /* clear overlay */
  /*****************/

  ImageAssign(&Ovl, OvlByteAddr(0,0), DispGetColumns, DispGetRows, OvlGetPitch);
  set(&Ovl,0);

  /***************************/
  /* Draw Screen and Overlay */
  /***************************/

  /* set background color */
  ImageAssign(&Ovl, OvlByteAddr(0, 0), DispGetColumns, DispGetRows, OvlGetPitch);
  set(&Ovl, COLGREY);

  /* open screen size */
  ImageAssign(&Ovl, OvlByteAddr(ScrX, ScrY), ScrDx, ScrDy, OvlGetPitch);
  set(&Ovl, 0);
  Low(&Ovl);
  
  /* Company */
  ImageAssign(&Ovl, OvlByteAddr(ScrX, 20), 400, 16, OvlGetPitch);
  chprint1(CompName,&Ovl, 2, 2, COLBLUE);
  ChangeColor(&Ovl,0,0,COLGREY);

  /* Version */
  ImageAssign(&Ovl, OvlBitAddr(ScrX + ScrDx - 8*(sizeof(PrgName)-1), 28), 8*(sizeof(PrgName)-1), 8, OvlGetPitch);
  chprint1(PrgName,&Ovl, 1, 1, COLBLUE);
  ChangeColor(&Ovl,0,0,COLGREY);

  /* Time */
  ImageAssign(&Par->Time, OvlBitAddr(ScrX + ScrDx - 64, DispGetRows - 26), 64, 8, OvlGetPitch);

  /* live image with overlay */
  set_ovlmask(255);
  vmode(vmOvlLive);

  /* now refer always to screen position */
  ScrSetLogPage(ScrByteAddr(ScrX, ScrY));
  OvlSetLogPage(OvlByteAddr(ScrX, ScrY));
}

/*************************************************************************************************************************/

void IniOverlayLUT(void)
{
  I32 i;
  I32 color;
  U32 table[256][3];
  volatile U32 *lut;


  // ini table with black (red=0 green=0 blue=0)
  for(i=0; i<256 ; i++)
  {	
    table[i][0]=0;	// RED		 
    table[i][1]=0;  // GREEN
    table[i][2]=0;	// BLUE		 
  }

  /******************************************************/
  /* Define Main Color (31 colors [8, 16, 24, ... 248]) */
  /******************************************************/

  color = COLBLACK;
  table[color][0] =  70;  // RED
  table[color][1] =  70;  // GREEN
  table[color][2] =  70;  // BLUE
   
  color = COLWHITE;
  table[color][0] = 240;  // RED
  table[color][1] = 240;  // GREEN
  table[color][2] = 240;  // BLUE
   
  color = COLBLUE;
  table[color][0] =  16;  // RED
  table[color][1] =   5;  // GREEN
  table[color][2] =  73;  // BLUE
   
  color = COLRED;
  table[color][0] = 250;  // RED
  table[color][1] =   0;  // GREEN
  table[color][2] =   0;  // BLUE
   
  color = COLGREEN;
  table[color][0] =  47;  // RED
  table[color][1] = 200;  // GREEN
  table[color][2] =  47;  // BLUE
   
  color = COLYELLOW;
  table[color][0] = 250;  // RED
  table[color][1] = 250;  // GREEN
  table[color][2] =   0;  // BLUE
   
  color = COLGREY;
  table[color][0] = 210;  // RED
  table[color][1] = 210;  // GREEN
  table[color][2] = 210;  // BLUE

  color = COLCYAN;
  table[color][0] =  47;  // RED
  table[color][1] = 191;  // GREEN
  table[color][2] = 191;  // BLUE

  color = COLMAGE;
  table[color][0] = 200;  // RED
  table[color][1] =  40;  // GREEN
  table[color][2] = 200;  // BLUE

  color = COLDARK;
  table[color][0] =  47;  // RED
  table[color][1] =  47;  // GREEN
  table[color][2] =  47;  // BLUE


  /******************************************************/
  /* programm overlay LUT                               */
  /******************************************************/

  lut = (volatile unsigned int *)VIRTX_OLUT;

  for(i=0; i<256 ; i++)
  {				 
    *lut++ = 
           ( (table[i][0])     ) +  /* RED   */
           ( (table[i][1]) << 8) +  /* GREEN */
           ( (table[i][2]) <<16) ;  /* BLUE  */
  }
}

/***************************************************************************************************************************/

void CapturePosition(int x, int y)
{
  image area;


  /* define image variable at the image page */
  ImageAssign(&area, ScrGetCaptPage, ScrGetColumns, ScrGetRows, ScrGetPitch);

  if (x < 0) x=0;
  if (x > DispGetColumns-ScrGetColumns) x=DispGetColumns-ScrGetColumns;

  x=(x/16)*16;

  if (y < 0) y=0;
  if (y > DispGetRows-ScrGetRows) y=DispGetRows-ScrGetRows;

  ScrSetCaptPage(ScrGetDispPage + x + y*ScrGetPitch);
   
  /* delete drawings from old position */
  set(&area,0);

  tpict(); // get new image
}

/***************************************************************************************************************************/

void IniParameter (Parameter *Par)
{
  I32 ReadError=0;

  // Read Parameter from File
  ReadError = ReadParameter(Par);

  // no new parameter file on flash ?
  if (ReadError)
  {
    /* shutter time in us */
    Par->StartShutter     = 10000;

    Par->Threshold        =    18;

    Par->ShowDetection    =     0;

    Par->SearchX          =    70;
    Par->SearchY          =   300;
    Par->SearchDx         =   500;
    Par->SearchDy         =    90;
  }
}

/*************************************************************************************************************************/

void IniCameraBack (Parameter *Par)

  {
  I32   DisplayPage;
  image Img;

  /* clear outputs */
  outPLC(0);

  /* Position of the image */
  DisplayPage = getvar(DISP_START);
  
  ScrSetPhysPage(DisplayPage);
  ScrSetLogPage (DisplayPage);
  OvlSetLogPage (DisplayPage);

  /* clear overlay */
  ImageAssign(&Img, OvlByteAddr(0, 0), DispGetColumns, DispGetRows, ScrGetPitch);
  set(&Img,0);

  /* clear image */
  ImageAssign(&Img, ScrByteAddr(0, 0), DispGetColumns, DispGetRows, ScrGetPitch);
  set(&Img,0);

  vmode(vmLive);
  }

/***************************************************************************************************************************/

void DisplayResults (Parameter *Par)
{
  char string[80];

  sprint(string,"%6dms", Par->Ms); 
  chprint1(string, &Par->Time, 1, 1, COLBLUE);
  ChangeColor(&Par->Time,0,0,COLGREY);
}

/***************************************************************************************************************************/

int ReadParameter (Parameter *Par)
{
  return( ReadFile("fd:/pat.dat", (char *)(Par), sizeof(Parameter)) );
}

/***************************************************************************************************************************/

int StoreParameter (Parameter *Par)
{
  if( ((I32)fremain()) < (sizeof(Parameter)+64) )	// +64 for header and savity
  {
    U32  res;
    FILE *fp;

    // Pack Flash
    fp = io_fopen("fd:/","c");
    io_ioctl(fp,IO_PACK, &res);
    io_fclose(fp);
  }

  return( StoreFile("fd:/pat.dat", (char *)(Par), sizeof(Parameter)) );
}

/*************************************************************************************************************************/

void DisplayMenu (Parameter *Par)

{
  int response;

  do
  {
    print("\n"); 
    print("--------------------------------------------------------------------------------\n");
    print("l:   live-picture\n"); 
    print("s:   change shutter time\n"); 
    print("t:   change threshold\n"); 
    print("a:   change search area\n"); 
    print("x:   display results\n"); 
    print("\n"); 
    print("q:   quit Menu\n"); 
    print("--------------------------------------------------------------------------------\n");

    response = ReadKeyboard(1);

    switch(response)
    {
      case 'l' : 
                 LiveImage(Par);
                 break;
      case 's' : 
                 // void ChangeValue (I32 *ParValue, I32 Min, I32 Max, I32 Step, U8 *text)
                 ChangeValue(&Par->StartShutter, 5, 100000, 1, "Shutter Time");
                 break;

      case 't' : 
                 // void ChangeValue (I32 *ParValue, I32 Min, I32 Max, I32 Step, U8 *text)
                 ChangeValue(&Par->Threshold, 1, 25, 1, "Threshold");
                 break;

      case 'a' : 
                 SelectArea(&Par->SearchX, &Par->SearchY, &Par->SearchDx, &Par->SearchDy, 0, ScrGetColumns, 0, ScrGetRows);
                 break;

      case 'x' : 
                 InvertValue(&Par->ShowDetection, "Show Detection");
                 break;

    }
  }
  while(response != 'q');

  SaveNewParameter(Par);
  
  print("RUN mode\n"); 
}

/*************************************************************************************************************************/

void InvertValue (I32 *ParValue, I8 *text)

{
  print("%s = ", text); 

  if(*ParValue) 
  {
    *ParValue=0;
    print("OFF\n"); 
  }
  else
  {
    *ParValue=1;
    print("ON\n"); 
  }

  print("--------------------------------------------------------------------------------\n");
}

/*************************************************************************************************************************/

void ChangeValue (I32 *ParValue, I32 Min, I32 Max, I32 Step, I8 *text)

{
  I32  value;
  I32  response=0;


  print("Select %s [%d - %d]\n", text, Min, Max);
  print("\n");
  print("Cursor Right: Increase Value\n");
  print("Cursor Left:  Decrease Value\n");
  print("ENTER:        Take New Value\n");
  print("ESC:          Take Old Value\n");
  print("\n");

  // start value
  value = *ParValue;

  // change value
  do
  {
    if(response == KEY_RIGHT) {value+=Step; if(value > Max) value=Max;}
    if(response == KEY_LEFT ) {value-=Step; if(value < Min) value=Min;}

    print("%s = %d  [%d - %d]\n", text, value, Min, Max); 

    response=ReadKeyboard(1);
  }
  while( (response != KEY_ESC) && (response != KEY_ENTER) );

  // take new value
  if (response == KEY_ENTER)  *ParValue = value;

  print("%s = %d\n", text, *ParValue); 

  print("--------------------------------------------------------------------------------\n");
}

/*************************************************************************************************************************/

void SaveNewParameter(Parameter *Par)

{
  char res;

  print("\n");
  print("Save New Parameters (y/n)?\n");

  do 
  {
    res = ReadKeyboard(1);
  
    if (res == 'y')
    {
      print("Saving New Parameters. Please Wait ...\n");
      StoreParameter(Par); 
      print("successful\n\n");
    }

    if (res == 'n')
    {
      print("Cancel Saving.\n");
    }
  }
  while(res!='y' && res!='n');
}

/***************************************************************************************************************************/

void LiveImage (Parameter *Par)
{
  shutter(Par->StartShutter);
  vmode(vmOvlLive);
}

/*************************************************************************************************************************/

void High (image *a)

{
  image area;
  I32   dx, dy;
  
  ImageAssign(&area, a->st-2L-2L*((long)a->pitch), a->dx+4, a->dy+4, a->pitch);

  dx = area.dx-1;
  dy = area.dy-1;
  
  lined(&area,    0,    0,   dx,    0, COLWHITE);
  lined(&area,    1,    1, dx-1,    1, COLWHITE);
  lined(&area,    0,    0,    0,   dy, COLWHITE);
  lined(&area,    1,    1,    1, dy-1, COLWHITE);

  lined(&area,    1,   dy,   dx,   dy, COLBLACK);
  lined(&area,    2, dy-1, dx-1, dy-1, COLBLACK);
  lined(&area,   dx,    1,   dx,   dy, COLBLACK);
  lined(&area, dx-1,    2, dx-1, dy-1, COLBLACK);
}

/*************************************************************************************************************************/

void Low (image *a)

{
  image area;
  I32   dx, dy;
  
  ImageAssign(&area, a->st-2L-2L*((long)a->pitch), a->dx+4, a->dy+4, a->pitch);
  
  dx = area.dx-1;
  dy = area.dy-1;

  lined(&area,    0,    0,   dx,    0, COLBLACK);
  lined(&area,    1,    1, dx-1,    1, COLBLACK);
  lined(&area,    0,    0,    0,   dy, COLBLACK);
  lined(&area,    1,    1,    1, dy-1, COLBLACK);

  lined(&area,    1,   dy,   dx,   dy, COLWHITE);
  lined(&area,    2, dy-1, dx-1, dy-1, COLWHITE);
  lined(&area,   dx,    1,   dx,   dy, COLWHITE);
  lined(&area, dx-1,    2, dx-1, dy-1, COLWHITE);
}

/***************************************************************************************************************************/

I32 ReadFile (char *name, char *data, I32 bytes)

{
  I32  res, error=0;
  FILE *fp;

  /* read datas from file */
  /************************/
  fp = io_fopen(name,"r");;
  if(fp)
  {
    res=io_read(fp, data, bytes);

    if (res == bytes)
    {
      /* success */
      error=0;
    }
    else
    {
      /* can't read from file */
      error=-2;
    }

    io_fclose(fp);
  }
  else
  {
    /* can't open file for reading */
    error=-1;
  }

  return(error);
}

/***************************************************************************************************************************/

I32 StoreFile (char *name, char *data, I32 bytes)

{
  I32  res, error=0;
  FILE *fp;

  /* delete old file if it exist */
  /*******************************/
  fp = io_fopen(name,"c");
  if(fp)
    {
    io_ioctl(fp, IO_DEL, NULL);
    io_fclose(fp);
    }
  else
    {
    /* no old File to delete */
    }


  /* write file */
  /**************/
  fp = io_fopen(name,"w");
  if(fp)
  {
    res=io_write(fp, data, bytes);

    if (res >= 0)  
    {
      /* success */
      error=0;
    }
    else
    {
      /* can't write to file */
      error=-2;
    }
    
    io_fclose(fp);
  }
  else
  {
    /* can't open file for writing */
    error=-1;
  }
  
  return(error);
}

/*************************************************************************************************************************/

void ChangeColor (image *area, I32 start, I32 end, I32 newcol)

{
  I32 i, j, dx, dy;
  long addr, pitch;
  U8 *restrict ppix;

  
  addr=area->st;
  dx=area->dx;
  dy=area->dy;
  pitch=(long)area->pitch;

  for (j=0; j<dy; j++)
  {
    ppix=(U8 *)(addr+(pitch*(long)j));

    for (i=0; i<dx; i++)
    {
      if ( (*ppix>=start) && (*ppix<=end) )
      {
        *ppix=newcol;
      }
      ppix++;
    }
  }
}

/*************************************************************************************************************************/

void SelectArea (I32 *px, I32 *py, I32 *pdx, I32 *pdy, I32 xmin, I32 xmax, I32 ymin, I32 ymax)

  {
  I32   x, y, dx, dy, response, increment=10, mode=0, smallstep, bigstep;
  image Ovl, Text;
  char  textstr[40];

  print("\n");
  print("Press F1     step size\n");
  print("Press F2     toggle between Position and Size\n");
  print("Press CURSOR to change values\n");
  print("Press RETURN to accept values\n");

  x  = *px;
  y  = *py;
  dx = *pdx;
  dy = *pdy;

  bigstep   = 10;
  smallstep =  1;
  
  increment = bigstep;

  ImageAssign(&Ovl,OvlBitAddr(x,y), dx, dy, OvlGetPitch);  

  /* select size */
  do
    {
    framez(&Ovl);

    ImageAssign(&Text, OvlBitAddr(x, y-10), dx, 8, OvlGetPitch);  
    sprint(textstr,"X=%d Y=%d DX=%d DY=%d", x, y, dx, dy); 
    chprint1(textstr,&Text,1,1,1);

    response = ReadKeyboard(1);
    switch(response)
        {
        case KEY_DOWN : if(mode) {dy+=increment; if(y+dy > ymax) dy=ymax- y;}
                        else     { y+=increment; if(y+dy > ymax)  y=ymax-dy;} break;
        case KEY_RIGHT: if(mode) {dx+=increment; if(x+dx > xmax) dx=xmax- x;}
                        else     { x+=increment; if(x+dx > xmax)  x=xmax-dx;} break;
        case KEY_UP   : if(mode) {dy-=increment; if(dy <    2) dy=2;} 
                        else     { y-=increment; if(y  < ymin)  y=ymin;} break;
        case KEY_LEFT : if(mode) {dx-=increment; if(dx <    2) dx=2;} 
                        else     { x-=increment; if(x  < xmin)  x=xmin;} break;

        case KEY_F1   : if(increment==bigstep) increment=smallstep; else increment=bigstep; break;
        case KEY_F2   : if(mode     == 1) mode =0; else mode = 1; break;
        }

    chprint1(textstr,&Text,1,1,0);

    framez(&Ovl);
    ImageAssign(&Ovl,OvlBitAddr(x,y), dx, dy, OvlGetPitch);  
    }
  while(response!=KEY_ENTER && response!=KEY_ESC);

  if (response==KEY_ENTER)
    {
    *px  =  x;
    *py  =  y;
    *pdx = dx;
    *pdy = dy;
    }
  }

/***************************************************************************************************************************/

#define TimeOut 10 /* ms */

I32 ReadKeyboard (I32 wait)
{
  I32 i, ch1, ch2, ch3, ch4, src=-1;


  if(wait || kbready() || rbready()) 
  {

  do
  {
    if (kbready()) {src=1; break;}
    if (rbready()) {src=0; break;}
  }
  while(src<0);

  if(src) ch1 = kbdrcv();
  else    ch1 = rs232rcv();

  if (ch1 != 0x1b)
     if(ch1 == 0x0d || ch1 == 0x0a)
       return(KEY_ENTER);
     else
       return(ch1);
       // return(KEY_INVALID); NEW: this is the old line

  ch1 = KEY_ESC;

  for (i=0; i<TimeOut; i++)
  {
    time_delay(1);
    if(src) {if (kbready()) break;}
    else    {if ( kbhit() ) break;}
  }
  if (i==TimeOut) return(ch1);  /* we've been waiting for too long */

  if(src) ch2 = kbdrcv();
  else    ch2 = rs232rcv();

  for (i=0; i<TimeOut; i++)
  {
    time_delay(1);
    if(src) {if (kbready()) break;}
    else    {if ( kbhit() ) break;}
  }
  if (i==TimeOut) return(ch1);  /* we've been waiting for too long */

  if(src) ch3 = kbdrcv();
  else    ch3 = rs232rcv();


  if (ch2 == 0x4f)   /* 2/3 of a F1 or F2 received? */
  {

    if ((ch3 == 0x71) || (ch3 == 0x50))   /* F1 received? */
       return(KEY_F1);
    
    if ((ch3 == 0x72) || (ch3 == 0x51))   /* F2 received? */
       return(KEY_F2);
    
    return(KEY_INVALID);
  }

  if (ch2 == 0x5B)   /* 2/3 of a Cursor received? */
  {
    if (ch3 == 0x31) 
    {
      for (i=0; i<TimeOut; i++)
      {
        time_delay(1);
        if(src) {if (kbready()) break;}
        else    {if ( kbhit() ) break;}
      }
      if (i==TimeOut) return(ch1);  /* we've been waiting for too long */

      if(src) ch4 = kbdrcv();
      else    ch4 = rs232rcv();

      if (ch4 == 0x31)   /* F1 received? */
       return(KEY_F1);
    
      if (ch4 == 0x32)   /* F2 received? */
       return(KEY_F2);
    }
    else
    {
      switch (ch3)
    {
      case 0x41 :
      case 0x42 :
      case 0x43 :
      case 0x44 : return(ch3 - 61);
      default   : return(KEY_INVALID);
    } 
     }
  }
  return(KEY_INVALID);
  }
  else
  {
  return(KEY_INVALID);
  }
}

/***************************************************************************************************************************/

#define MAXOBJECT     200
#define MAXLENGTH  100000L
#define ACCURACY        1

/***************************************************************************************************************************/

int MeasureObject (Parameter *Par)
{
  I32   i, x, y, dx, dy, color, nobj, MarkerX, MarkerY, MarkerSize, BlackX, BlackY, MarkerDistance;
  I32   BlobNr[MAXOBJECT];
  char  string[80];
  image area, Marker;
  ftr   f[MAXOBJECT];


  // SendMessage(Par, " ", COLGREY);
  
  Par->Error     =  GOOD;
  MarkerSize     =     7;
  MarkerDistance = 84+42;

 
  /********************************************************/
  /*  find markers                                        */
  /********************************************************/
 
  color   = 0;

  nobj = FindMarkers(Par, Par->Threshold, f, Par->ShowDetection);

  // Generate Blob Nr
  for(i=0; i<nobj; i++) BlobNr[i]=i;

  /*
  for(i=0; i<nobj; i++) 
  {
    print("Blob[%2d]=%2d:  x=%3d y=%3d area=%d\n", i, BlobNr[i], f[BlobNr[i]].x_center, f[BlobNr[i]].y_center, f[BlobNr[i]].area);
  }
  print("\n");
  */

  // delete wrong colors and if blob connects border
  for(i=0; i<nobj; i++) 
  {
    if(f[i].color !=  color) f[i].area=0;
    if(f[i].x_min ==      0) f[i].area=0;
    if(f[i].x_max ==   dx-1) f[i].area=0;
    if(f[i].y_min ==      0) f[i].area=0;
    if(f[i].y_max ==   dy-1) f[i].area=0;
  }

  if (nobj >= MAX_MARKER)
  {
  // sort blobs by size (smallest blobs at the beginning)
  SortStructure(&BlobNr[0], nobj, &f[0].area, sizeof(ftr));

  // sort biggest 4 blobs by x-center (smallest x-center at the beginning)
  SortStructure(&BlobNr[nobj-MAX_MARKER], MAX_MARKER, &f[0].x_center, sizeof(ftr));
  }


  /********************************************************/
  /*  draw results                                        */
  /********************************************************/

  x  =  79;
  y  = ScrGetRows-96;
  dx = 482;
  dy =  86;
  ImageAssign(&area, OvlByteAddr(x, y), dx, dy, OvlGetPitch);
  color=COLGREY;
  set(&area, color);
  High(&area);

  for(i=0; i<4; i++)
  {
    x  =  79 + 10 + i*(84+42);
    y  = ScrGetRows-90 + 10;
    dx =  84;
    dy =  60;

    if (i==0) {color=COLDARK;	sprint(string,"Black");}
    if (i==1) {color=COLCYAN;  	sprint(string,"Cyan");}
    if (i==2) {color=COLMAGE;	sprint(string,"Magenta");}
    if (i==3) {color=COLYELLOW;	sprint(string,"Yellow");}
    
    ImageAssign(&area, OvlByteAddr(x, y), dx, dy, OvlGetPitch);

    framed (&area, color);
    markerd(&area, color);

    ImageAssign(&area, OvlByteAddr(x, y-10), dx, 8, OvlGetPitch);
    chprint1(string, &area, 1, 1, color);
    ChangeColor(&area, 0, 0, COLGREY);

    
    /********************************************************/
    /*  draw marker position                                */
    /********************************************************/

    if (nobj >= MAX_MARKER)
    {
      // show blobs
  
      // print("Blob[%2d]=%2d:  x=%3d y=%3d area=%6d (nobj=%3d)\n", i, BlobNr[i], f[BlobNr[i]].x_center, f[BlobNr[i]].y_center, f[BlobNr[i]].area, nobj);

      BlackX = f[BlobNr[nobj-MAX_MARKER]].x_center;
      BlackY = f[BlobNr[nobj-MAX_MARKER]].y_center;

      MarkerX = x +	(dx - MarkerSize)/2 + f[BlobNr[nobj-MAX_MARKER+i]].x_center - BlackX - i * MarkerDistance;
      MarkerY = y + (dy - MarkerSize)/2 + f[BlobNr[nobj-MAX_MARKER+i]].y_center - BlackY;

      if (MarkerX>x && MarkerX<x+dx-MarkerSize && MarkerY>y && MarkerY<y+dy-MarkerSize)
      {
        ImageAssign(&Marker, OvlByteAddr(MarkerX, MarkerY), MarkerSize, MarkerSize, OvlGetPitch);
        set(&Marker, color);
      }
    }
  }

  return(0);
}

/************************************************************************************************************************/

I32 FindMarkers (Parameter *Par, I32 threshold, ftr *f, I32 ShowDetection)
{
    I32 kernal;
    I32 x, y, dx, dy;
    I32 NewScreen;
    image areaBig, tempBig, temp, area;
    U16 *rlc;
    U16 *slc;
    int  objects;

    kernal = 40;

    /* allocate new screen page */
    NewScreen=(int)DRAMScreenMalloc(); 

    x  =  Par->SearchX;
    y  =  Par->SearchY;
    dx =  Par->SearchDx;
    dy =  Par->SearchDy;

  
    /* Search area */
    ImageAssign(&area, ScrByteAddr(x, y), dx, dy, ScrGetPitch);
    ImageAssign(&areaBig, ScrByteAddr(x-kernal, y-kernal), dx+2*kernal, dy+2*kernal, ScrGetPitch);

    ImageAssign(&temp, NewScreen+kernal+kernal*ScrGetPitch, dx, dy, ScrGetPitch);
    ImageAssign(&tempBig, NewScreen, dx+2*kernal, dy+2*kernal, ScrGetPitch);


    maskx(&areaBig, &tempBig, kernal, kernal, threshold);


    if(ShowDetection)
    {
      copy(&temp, &area);
      binarize(&area, &area, 4, 0, 255);
      framed(&area, 125);
    }


  /* allocate DRAM Memory */
  rlc = (U16 *) sysmalloc((MAXLENGTH * sizeof(U16)+1) / sizeof(U32), MDATA);
  if (rlc == NULL) {pstr("DRAM Memory overrun\n"); return(-1);}


  /* create RLC */
  slc = rlcmk(&temp, 4, rlc, MAXLENGTH-32);
  if (slc == NULL) {pstr("RLC overrun\n"); sysfree(rlc); return(-1);}

  
  /* object labelling */
  if(sgmt(rlc, slc)==0L) {pstr("object number overrun\n"); sysfree(rlc); return(-1);}

  /* feature extraction */
  objects = rl_ftr2(rlc, f, MAXOBJECT);

  /* free allocation */
  sysfree(rlc);  

  /* release screen allocation */
  DRAMPgFree(NewScreen);

  return(objects);
}

/***************************************************************************************************************************/

void SortStructure(I32 *StructNr, I32 number, U32 *SortValue, I32 StructureSize)
{
  I32 i, j, temp;
  U32  value1, value2;

  /* first is smallest number, last is highest number */

  StructureSize = StructureSize / sizeof(U32);

  for (i=0; i<number-1; i++)
  {
    for (j=0; j<number-1-i; j++)
    {
      value1 = SortValue[StructNr[  j]*StructureSize];
      value2 = SortValue[StructNr[1+j]*StructureSize];

      // print("value1=%d value2=%d\n", value1, value2);

      if ( value1 > value2 )
      {
        temp             = StructNr[  j];
        StructNr[  j]  = StructNr[1+j];
        StructNr[1+j]  = temp;
      }
    }
  }
}

/***************************************************************************************************************************/

